#pragma once

#ifdef DOXYGEN
#define BEGIN_ALPAQA_CASADI_LOADER_NAMESPACE
#define END_ALPAQA_CASADI_LOADER_NAMESPACE
#elif ALPAQA_WITH_EXTERNAL_CASADI
#define BEGIN_ALPAQA_CASADI_LOADER_NAMESPACE inline namespace external {
#define END_ALPAQA_CASADI_LOADER_NAMESPACE }
#else
#define BEGIN_ALPAQA_CASADI_LOADER_NAMESPACE inline namespace builtin {
#define END_ALPAQA_CASADI_LOADER_NAMESPACE }
#endif
