#pragma once

#include <guanaqo/linalg/sparsity.hpp>

namespace alpaqa::sparsity {
using namespace guanaqo::linalg::sparsity;
} // namespace alpaqa::sparsity

namespace alpaqa {
using sparsity::Sparsity;
} // namespace alpaqa
